@@ -4,8 +4,9 @@
 /// Interface definition for Soapy SDR devices.
 ///
 /// \copyright
-/// Copyright (c) 2014-2018 Josh Blum
+/// Copyright (c) 2014-2019 Josh Blum
 /// Copyright (c) 2016-2016 Bastille Networks
+///                    2021 Nicholas Corgan
 /// SPDX-License-Identifier: BSL-1.0
 ///
 
@@ -93,6 +94,16 @@ public:
      */
     static std::vector<Device *> make(const KwargsList &argsList);
 
+    /*!
+     * Create a list of devices from a list of construction arguments.
+     * This is a convenience call to parallelize device construction,
+     * and is fundamentally a parallel for loop of make(args).
+     *
+     * \param argsList a list of device arguments per each device
+     * \return a list of device pointers per each specified argument
+     */
+    static std::vector<Device *> make(const std::vector<std::string> &argsList);
+
     /*!
      * Unmake or release a list of device handles.
      * This is a convenience call to parallelize device destruction,
@@ -619,6 +630,30 @@ public:
      */
     virtual std::complex<double> getIQBalance(const int direction, const size_t channel) const;
 
+    /*!
+     * Does the device support automatic frontend IQ balance correction?
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \return true if IQ balance corrections are supported
+     */
+    virtual bool hasIQBalanceMode(const int direction, const size_t channel) const;
+
+    /*!
+     * Set the automatic frontend IQ balance correction.
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \param automatic true for automatic correction
+     */
+    virtual void setIQBalanceMode(const int direction, const size_t channel, const bool automatic);
+
+    /*!
+     * Set the automatic IQ balance corrections mode.
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \return true for automatic correction
+     */
+    virtual bool getIQBalanceMode(const int direction, const size_t channel) const;
+
     /*!
      * Does the device support frontend frequency correction?
      * \param direction the channel direction RX or TX
@@ -932,6 +967,24 @@ public:
      */
     virtual RangeList getMasterClockRates(void) const;
 
+    /*!
+     * Set the reference clock rate of the device.
+     * \param rate the clock rate in Hz
+     */
+    virtual void setReferenceClockRate(const double rate);
+
+    /*!
+     * Get the reference clock rate of the device.
+     * \return the clock rate in Hz
+     */
+    virtual double getReferenceClockRate(void) const;
+
+    /*!
+     * Get the range of available reference clock rates.
+     * \return a list of clock rate ranges in Hz
+     */
+    virtual RangeList getReferenceClockRates(void) const;
+
     /*!
      * Get the list of available clock sources.
      * \return a list of clock source names
@@ -1033,6 +1086,15 @@ public:
      */
     virtual std::string readSensor(const std::string &key) const;
 
+    /*!
+     * Readback a global sensor given the name.
+     * \tparam Type the return type for the sensor value
+     * \param key the ID name of an available sensor
+     * \return the current value of the sensor as the specified type
+     */
+    template <typename Type>
+    Type readSensor(const std::string &key) const;
+
     /*!
      * List the available channel readback sensors.
      * A sensor can represent a reference lock, RSSI, temperature.
@@ -1063,6 +1125,17 @@ public:
      */
     virtual std::string readSensor(const int direction, const size_t channel, const std::string &key) const;
 
+    /*!
+     * Readback a channel sensor given the name.
+     * \tparam Type the return type for the sensor value
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \param key the ID name of an available sensor
+     * \return the current value of the sensor as the specified type
+     */
+    template <typename Type>
+    Type readSensor(const int direction, const size_t channel, const std::string &key) const;
+
     /*******************************************************************
      * Register API
      ******************************************************************/
@@ -1146,6 +1219,15 @@ public:
      */
     virtual void writeSetting(const std::string &key, const std::string &value);
 
+    /*!
+     * Write a setting with an arbitrary value type.
+     * \tparam Type the data type of the value
+     * \param key the setting identifier
+     * \param value the setting value
+     */
+    template <typename Type>
+    void writeSetting(const std::string &key, const Type &value);
+
     /*!
      * Read an arbitrary setting on the device.
      * \param key the setting identifier
@@ -1153,6 +1235,15 @@ public:
      */
     virtual std::string readSetting(const std::string &key) const;
 
+    /*!
+     * Read an arbitrary setting on the device.
+     * \tparam Type the return type for the sensor value
+     * \param key the setting identifier
+     * \return the setting value
+     */
+    template <typename Type>
+    Type readSetting(const std::string &key);
+
     /*!
      * Describe the allowed keys and values used for channel settings.
      * \param direction the channel direction RX or TX
@@ -1171,6 +1262,17 @@ public:
      */
     virtual void writeSetting(const int direction, const size_t channel, const std::string &key, const std::string &value);
 
+    /*!
+     * Write an arbitrary channel setting on the device.
+     * \tparam Type the data type of the value
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \param key the setting identifier
+     * \param value the setting value
+     */
+    template <typename Type>
+    void writeSetting(const int direction, const size_t channel, const std::string &key, const Type &value);
+
     /*!
      * Read an arbitrary channel setting on the device.
      * \param direction the channel direction RX or TX
@@ -1180,6 +1282,17 @@ public:
      */
     virtual std::string readSetting(const int direction, const size_t channel, const std::string &key) const;
 
+    /*!
+     * Read an arbitrary channel setting on the device.
+     * \tparam Type the return type for the sensor value
+     * \param direction the channel direction RX or TX
+     * \param channel an available channel on the device
+     * \param key the setting identifier
+     * \return the setting value
+     */
+    template <typename Type>
+    Type readSetting(const int direction, const size_t channel, const std::string &key);
+
     /*******************************************************************
      * GPIO API
      ******************************************************************/
@@ -1308,6 +1421,53 @@ public:
      */
     virtual std::string readUART(const std::string &which, const long timeoutUs = 100000) const;
 
+    /*******************************************************************
+     * Native Access API
+     ******************************************************************/
+
+    /*!
+     * A handle to the native device used by the driver.
+     * The implementation may return a null value if it does not support
+     * or does not wish to provide access to the native handle.
+     * \return a handle to the native device or null
+     */
+    virtual void* getNativeDeviceHandle(void) const;
 };
 
-};
+}
+
+template <typename Type>
+Type SoapySDR::Device::readSensor(const std::string &key) const
+{
+    return SoapySDR::StringToSetting<Type>(this->readSensor(key));
+}
+
+template <typename Type>
+Type SoapySDR::Device::readSensor(const int direction, const size_t channel, const std::string &key) const
+{
+    return SoapySDR::StringToSetting<Type>(this->readSensor(direction, channel, key));
+}
+
+template <typename Type>
+void SoapySDR::Device::writeSetting(const std::string &key, const Type &value)
+{
+    this->writeSetting(key, SoapySDR::SettingToString(value));
+}
+
+template <typename Type>
+Type SoapySDR::Device::readSetting(const std::string &key)
+{
+    return SoapySDR::StringToSetting<Type>(this->readSetting(key));
+}
+
+template <typename Type>
+void SoapySDR::Device::writeSetting(const int direction, const size_t channel, const std::string &key, const Type &value)
+{
+    this->writeSetting(direction, channel, key, SoapySDR::SettingToString(value));
+}
+
+template <typename Type>
+Type SoapySDR::Device::readSetting(const int direction, const size_t channel, const std::string &key)
+{
+    return SoapySDR::StringToSetting<Type>(this->readSetting(direction, channel, key));
+}
